// Poker Mavens Card Shuffling Simulator
// Copyright  2010-2016 Kent Briggs

unit Main;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls, AppEvnts, ComCtrls, VirtualTrees, System.UITypes;

type
  TForm1 = class(TForm)
    SaveDialog1: TSaveDialog;
    CardView: TVirtualStringTree;
    Panel2: TPanel;
    HoleCardPanel: TPanel;
    CommunityCardPanel: TPanel;
    HandPanel: TPanel;
    PageControl1: TPageControl;
    ShuffleTab: TTabSheet;
    Label1: TLabel;
    HandCount: TEdit;
    DealBtn: TButton;
    ClearBtn: TButton;
    SaveBtn: TButton;
    PokerTab: TTabSheet;
    Label6: TLabel;
    Label7: TLabel;
    Label8: TLabel;
    Label9: TLabel;
    PlayersBox: TComboBox;
    PlayerPosBox: TComboBox;
    DealtBox: TComboBox;
    PlayerPosStatsBtn: TButton;
    FindHandBox: TComboBox;
    FindNextBtn: TButton;
    CardTab: TTabSheet;
    Label3: TLabel;
    Label5: TLabel;
    Label4: TLabel;
    Label2: TLabel;
    CardBox: TComboBox;
    CardStatsBtn: TButton;
    RankBox: TComboBox;
    RankStatsBtn: TButton;
    SuitBox: TComboBox;
    SuitStatsBtn: TButton;
    DeckPosBox: TComboBox;
    DeckPosStatsBtn: TButton;
    procedure FormCreate(Sender: TObject);
    procedure DealBtnClick(Sender: TObject);
    procedure ClearBtnClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure SaveBtnClick(Sender: TObject);
    procedure DeckPosStatsBtnClick(Sender: TObject);
    procedure CardStatsBtnClick(Sender: TObject);
    procedure SuitStatsBtnClick(Sender: TObject);
    procedure RankStatsBtnClick(Sender: TObject);
    procedure PlayerPosStatsBtnClick(Sender: TObject);
    procedure PlayersBoxChange(Sender: TObject);
    procedure CardViewGetText(Sender: TBaseVirtualTree; Node: PVirtualNode;
      Column: TColumnIndex; TextType: TVSTTextType; var CellText: string);
    procedure CardViewPaintText(Sender: TBaseVirtualTree;
      const TargetCanvas: TCanvas; Node: PVirtualNode; Column: TColumnIndex;
      TextType: TVSTTextType);
    procedure PlayerPosBoxChange(Sender: TObject);
    procedure DealtBoxChange(Sender: TObject);
    procedure CardViewFocusChanged(Sender: TBaseVirtualTree; Node: PVirtualNode;
      Column: TColumnIndex);
    procedure FindNextBtnClick(Sender: TObject);
    procedure PageControl1Resize(Sender: TObject);
  private
    { Private declarations }
    procedure GetHandCards(i: integer; var H1, H2, B1, B2, B3, B4, B5: integer);
    procedure ShowHand(Node: PVirtualNode);
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation
uses PRNG, Stats, PokerHand;

var
  Cards: TCardDeck;
  CardList: TList;
  Players: integer = 10;

{$R *.dfm}

function GetCard(c: integer): string;
begin
  Result := CardFace[c] + CardSuit[c];
end;

procedure TForm1.ClearBtnClick(Sender: TObject);
var
  i: integer;
  Deck: PCardDeck;
begin
  CardView.RootNodeCount:=0;
  for i := CardList.Count - 1 downto 0 do
  begin
    Deck := CardList[i];
    Dispose(Deck);
  end;
  CardList.Clear;
end;

procedure TForm1.DealBtnClick(Sender: TObject);
var
  i, j, n: integer;
  Deck: PCardDeck;
begin
  n := StrToIntDef(HandCount.Text, 0);
  if n < 1 then
  begin
    MessageDlg('Invalid hand count', mtError, [mbOk], 0);
    Exit;
  end;
  Screen.Cursor := crHourGlass;
  for i := 1 to n do
  begin
    Prng_Shuffle(@Cards);
    New(Deck);
    for j := 1 to 52 do Deck^[j] := Cards[j];
    CardList.Add(Deck);
  end;
  CardView.RootNodeCount := CardList.Count;
  Screen.Cursor := crDefault;
end;

procedure TForm1.DealtBoxChange(Sender: TObject);
begin
  CardView.Invalidate;
  ShowHand(CardView.FocusedNode);
end;

procedure TForm1.FindNextBtnClick(Sender: TObject);
var
  Target, H1, H2, B1, B2, B3, B4, B5, Rank: integer;
  Node: PVirtualNode;
  Found: boolean;
begin
  if CardList.Count = 0 then
  begin
    MessageDlg('No hands dealt', mtError, [mbOk], 0);
    Exit;
  end;
  Target := FindHandBox.ItemIndex;
  Found := False;
  Node := CardView.GetFirstSelected;
  repeat
    if Node = Nil then Node := CardView.GetFirst else Node := CardView.GetNext(Node);
    if Node = Nil then Break;
    GetHandCards(Node.Index, H1, H2, B1, B2, B3, B4, B5);
    Rank := HandRank(H1, H2, B1, B2, B3, B4, B5);
    case Target of
      0: Found := Rank < $200000;  // high card
      1: Found := Rank < $300000;  // pair -
      2: Found := (Rank < $300000) and (Rank >= $200000);  // pair
      3: Found := Rank >= $200000; // pair +
      4: Found := Rank < $400000;  // two pair -
      5: Found := (Rank < $400000) and (Rank >= $300000);  // two pair
      6: Found := Rank >= $300000; // two pair +
      7: Found := Rank < $500000;  // three of a kind -
      8: Found := (Rank < $500000) and (Rank >= $400000);  // three of a kind
      9: Found := Rank >= $400000; // three of a kind +
      10: Found := Rank < $600000;  // straight -
      11: Found := (Rank < $600000) and (Rank >= $500000);  // straight
      12: Found := Rank >= $500000; // straight +
      13: Found := Rank < $700000;  // flush -
      14: Found := (Rank < $700000) and (Rank >= $600000);  // flush
      15: Found := Rank >= $600000; // flush +
      16: Found := Rank < $800000;  // full house -
      17: Found := (Rank < $800000) and (Rank >= $700000);  // full house
      18: Found := Rank >= $700000; // full house +
      19: Found := Rank < $900000;  // four of a kind -
      20: Found := (Rank < $900000) and (Rank >= $800000);  // four of a kind
      21: Found := Rank >= $800000; // four of a kind +
      22: Found := Rank < $9EDCBA;  // straight flush -
      23: Found := (Rank < $9EDCBA) and (Rank >= $900000);  // straight flush
      24: Found := Rank >= $900000; // straight flush +
      25: Found := Rank = $9EDCBA; // royal flush
    end;
  until Found;
  if not(Found) then MessageDlg('End of List', mtInformation, [mbOk], 0) else
  begin
    CardView.ScrollIntoView(Node, True, False);
    CardView.Selected[Node] := True;
    ShowHand(Node);
  end;
end;

procedure TForm1.FormCreate(Sender: TObject);
var i: integer;
begin
  for i := 1 to 52 do
  begin
    Cards[i] := i;  // initialize card deck
    DeckPosBox.Items.Add(IntToStr(i));  // fill position list
    CardBox.Items.Add(GetCard(i));  // fill card list;
    if (i - 1) mod 4 = 0 then RankBox.Items.Add(CardFace[i]);
  end;
  DeckPosBox.ItemIndex := 0;
  CardBox.ItemIndex := 0;
  RankBox.ItemIndex := 0;
  PlayersBoxChange(Nil);
  CardList := TList.Create;
  with PageControl1 do TabWidth := (ClientWidth - 10) div PageCount;
end;

procedure TForm1.FormDestroy(Sender: TObject);
var
  i: integer;
  Deck: PCardDeck;
begin
  for i := CardList.Count - 1 downto 0 do
  begin
    Deck := CardList[i];
    Dispose(Deck);
  end;
  CardList.Free;
end;

procedure TForm1.SaveBtnClick(Sender: TObject);
var
  FName,s : string;
  F: TextFile;
  i, j: integer;
  Deck: PCardDeck;
begin
  with SaveDialog1 do
  begin
    if Execute = False then Exit;
    FName := FileName;
  end;
  AssignFile(F, FName);
  Rewrite(F);
  for i := 0 to CardList.Count - 1 do
  begin
    Deck := CardList[i];
    s := GetCard(Deck^[1]);
    for j := 2 to 52 do s := s + ' ' + GetCard(Deck^[j]);
    WriteLn(F, s);
  end;
  CloseFile(F);
end;

procedure TForm1.SuitStatsBtnClick(Sender: TObject);
var
  i, j, n, cs1, cs2: integer;
  pt: string;
  StatsList: TStringList;
  PosCount: array[1..52] of integer;
  Deck: PCardDeck;
begin
  n := CardList.Count;
  if n = 0 then
  begin
    MessageDlg('No hands dealt', mtError, [mbOk], 0);
    Exit;
  end;
  StatsList := TStringList.Create;
  cs1 := SuitBox.ItemIndex;
  StatsList.Add('Suit Selected: ' + SuitBox.Text);
  StatsList.Add('Shuffles: ' + IntToStr(n));
  StatsList.Add('');
  StatsList.Add('Position = Count  (Percent) Avg = 25.0000%');
  StatsList.Add('');
  for i := 1 to 52 do PosCount[i] := 0;
  for i := 0 to n - 1 do
  begin
    Deck := CardList[i];
    for j := 1 to 52 do
    begin
      cs2 := (Deck^[j] - 1) mod 4;
      if cs1 = cs2 then Inc(PosCount[j]);
    end;
  end;
  for i := 1 to 52 do
  begin
    j := PosCount[i];
    pt := FloatToStrF(j * 100 / n, ffFixed, 18, 4);
    StatsList.Add(IntToStr(i) + ' = ' + IntToStr(j) + '  (' + pt + '%)');
  end;
  with StatsForm do
  begin
    Memo1.Lines.Text := StatsList.Text;
    ShowModal;
  end;
  StatsList.Free;
end;

procedure TForm1.PageControl1Resize(Sender: TObject);
begin
  with PageControl1 do TabWidth := (ClientWidth - 10) div PageCount;
end;

procedure TForm1.PlayerPosBoxChange(Sender: TObject);
begin
  CardView.Invalidate;
  ShowHand(CardView.FocusedNode);
end;

procedure TForm1.PlayerPosStatsBtnClick(Sender: TObject);
var
  i, n, Card1, Card2, Card3, Card4, Card5, Card6, Card7: integer;
  Rank, HC1, HC2, B1, B2, B3, B4, B5, Dealt: integer;
  HighCard, Pair, TwoPair, ThreeKind, Straight, Flush, FullHouse: integer;
  FourKind, StraightFlush, RoyalFlush, Error: integer;
  Rotate: boolean;
  s, pt: string;
  StatsList: TStringList;
  Deck: PCardDeck;
begin
  n := CardList.Count;
  if n = 0 then
  begin
    MessageDlg('No hands dealt', mtError, [mbOk], 0);
    Exit;
  end;
  Rotate := PlayerPosBox.ItemIndex >= Players;
  StatsList := TStringList.Create;
  StatsList.Add('Player position: ' + PlayerPosBox.Text);
  StatsList.Add('Cards dealt: ' + DealtBox.Text);
  StatsList.Add('Shuffles: ' + IntToStr(n));
  StatsList.Add('');
  Dealt := DealtBox.ItemIndex + 1;
  Card1 := Players - (PlayerPosBox.ItemIndex mod Players);
  Card2 := Card1 + Players;
  Card3 := (Players * 2) + 1;
  Card4 := Card3 + 1;
  Card5 := Card4 + 1;
  Card6 := Card5 + 1;
  Card7 := Card6 + 1;
  HighCard := 0;
  Pair := 0;
  TwoPair := 0;
  ThreeKind := 0;
  Straight := 0;
  Flush := 0;
  FullHouse := 0;
  FourKind := 0;
  StraightFlush := 0;
  RoyalFlush := 0;
  Error := 0;
  B1 := 0;
  B2 := 0;
  B3 := 0;
  B4 := 0;
  B5 := 0;
  for i := 0 to n - 1 do
  begin
    Deck := CardList[i];
    HC1 := Deck^[Card1];  // hole cards
    HC2 := Deck^[Card2];
    if Dealt > 1 then     // flop
    begin
      B1 := Deck^[Card3];
      B2 := Deck^[Card4];
      B3 := Deck^[Card5];
    end;
    if Dealt > 2 then B4 := Deck^[Card6];  // turn
    if Dealt > 3 then B5 := Deck^[Card7];  // river
    Rank := HandRank(HC1, HC2, B1, B2, B3, B4, B5);
    if Rank= $9EDCBA then Inc(RoyalFlush) else
    if Rank>=$900000 then Inc(StraightFlush) else
    if Rank>=$800000 then Inc(FourKind) else
    if Rank>=$700000 then Inc(FullHouse) else
    if Rank>=$600000 then Inc(Flush) else
    if Rank>=$500000 then Inc(Straight) else
    if Rank>=$400000 then Inc(ThreeKind) else
    if Rank>=$300000 then Inc(TwoPair) else
    if Rank>=$200000 then Inc(Pair) else
    if Rank>=$100000 then Inc(HighCard)
    else Inc(Error);
    if Rotate then
    begin
      Card1 := Card1 - 1;
      if Card1 < 1 then Card1 := Players;
      Card2 := Card1 + Players;
    end;
  end;
  pt := FloatToStrF(HighCard * 100 / n, ffFixed, 18, 4);
  StatsList.Add('High Card:       ' + IntToStr(HighCard) + '  (' + pt + '%)');
  pt := FloatToStrF(Pair * 100 / n, ffFixed, 18, 4);
  StatsList.Add('Pair:            ' + IntToStr(Pair) + '  (' + pt + '%)');
  if Dealt > 1 then
  begin
    pt := FloatToStrF(TwoPair * 100 / n, ffFixed, 18, 4);
    StatsList.Add('Two Pair:        ' + IntToStr(TwoPair) + '  (' + pt + '%)');
    pt := FloatToStrF(ThreeKind * 100 / n, ffFixed, 18, 4);
    StatsList.Add('Three of a Kind: ' + IntToStr(ThreeKind) + '  (' + pt + '%)');
    pt := FloatToStrF(Straight * 100 / n, ffFixed, 18, 4);
    StatsList.Add('Straight:        ' + IntToStr(Straight) + '  (' + pt + '%)');
    pt := FloatToStrF(Flush * 100 / n, ffFixed, 18, 4);
    StatsList.Add('Flush:           ' + IntToStr(Flush) + '  (' + pt + '%)');
    pt := FloatToStrF(FullHouse * 100 / n, ffFixed, 18, 4);
    StatsList.Add('Full House:      ' + IntToStr(FullHouse) + '  (' + pt + '%)');
    pt := FloatToStrF(FourKind * 100 / n, ffFixed, 18, 4);
    StatsList.Add('Four of a Kind:  ' + IntToStr(FourKind) + '  (' + pt + '%)');
    pt := FloatToStrF(StraightFlush * 100 / n, ffFixed, 18, 4);
    StatsList.Add('Straight Flush:  ' + IntToStr(StraightFlush) + '  (' + pt + '%)');
    pt := FloatToStrF(RoyalFlush * 100 / n, ffFixed, 18, 4);
    StatsList.Add('Royal Flush:     ' + IntToStr(RoyalFlush) + '  (' + pt + '%)');
  end;

  with StatsForm do
  begin
    Memo1.Lines.Text := StatsList.Text;
    ShowModal;
  end;
  StatsList.Free;
end;

procedure TForm1.PlayersBoxChange(Sender: TObject);
var i: integer;
begin
  Players := StrToInt(PlayersBox.Text);
  PlayerPosBox.Clear;
  PlayerPosBox.Items.Add('D');
  for i := 1 to Players - 1 do PlayerPosBox.Items.Add('D-' + IntToStr(i));
  PlayerPosBox.Items.Add('D  (rotate)');
  for i := 1 to Players - 1 do PlayerPosBox.Items.Add('D-' + IntToStr(i) + '  (rotate)');
  PlayerPosBox.ItemIndex := 0;
  CardView.Invalidate;
  ShowHand(CardView.FocusedNode);
end;

procedure TForm1.DeckPosStatsBtnClick(Sender: TObject);
const
  CardR = '23456789TJQKA';
  CardS = 'cdhs';
var
  p, i, j, n, CardNum, SuitNum, RankNum: integer;
  pt: string;
  StatsList: TStringList;
  SuitCount: array[1..4] of integer;
  RankCount: array[1..13] of integer;
  CardCount: array[1..52] of integer;
  Deck: PCardDeck;
begin
  n := CardList.Count;
  if n = 0 then
  begin
    MessageDlg('No hands dealt', mtError, [mbOk], 0);
    Exit;
  end;
  StatsList := TStringList.Create;
  StatsList.Add('Deck Position: ' + DeckPosBox.Text);
  StatsList.Add('Shuffles: ' + IntToStr(n));
  StatsList.Add('');
  StatsList.Add('Suit = Count  (Percent) Avg = 25.0000%');
  StatsList.Add('');

  p := DeckPosBox.ItemIndex + 1;
  for i := 1 to 4 do SuitCount[i] := 0;
  for i := 0 to n - 1 do
  begin
    Deck := CardList[i];
    SuitNum := ((Deck^[p] - 1) mod 4) + 1;
    Inc(SuitCount[SuitNum]);
  end;
  for i := 1 to 4 do
  begin
    j := SuitCount[i];
    pt := FloatToStrF(j * 100 / n, ffFixed, 18, 4);
    StatsList.Add(CardS[i] + ' = ' + IntToStr(j) + '  (' + pt + '%)');
  end;
  StatsList.Add('');
  StatsList.Add('Rank = Count  (Percent) Avg = 7.6923%');
  StatsList.Add('');

  for i := 1 to 13 do RankCount[i] := 0;
  for i := 0 to n - 1 do
  begin
    Deck := CardList[i];
    RankNum := ((Deck^[p] - 1) div 4) + 1;
    Inc(RankCount[RankNum]);
  end;
  for i := 1 to 13 do
  begin
    j := RankCount[i];
    pt := FloatToStrF(j * 100 / n, ffFixed, 18, 4);
    StatsList.Add(CardR[i] + ' = ' + IntToStr(j) + '  (' + pt + '%)');
  end;
  StatsList.Add('');
  StatsList.Add('Card = Count  (Percent) Avg = 1.9231%');
  StatsList.Add('');

  for i := 1 to 52 do CardCount[i] := 0;
  for i := 0 to n - 1 do
  begin
    Deck := CardList[i];
    CardNum := Deck^[p];
    Inc(CardCount[CardNum]);
  end;
  for i := 1 to 52 do
  begin
    j := CardCount[i];
    pt := FloatToStrF(j * 100 / n, ffFixed, 18, 4);
    StatsList.Add(GetCard(i) + ' = ' + IntToStr(j) + '  (' + pt + '%)');
  end;

  with StatsForm do
  begin
    Memo1.Lines.Text := StatsList.Text;
    ShowModal;
  end;
  StatsList.Free;
end;

procedure TForm1.RankStatsBtnClick(Sender: TObject);
var
  i, j, n, cr1, cr2: integer;
  pt: string;
  StatsList: TStringList;
  PosCount: array[1..52] of integer;
  Deck: PCardDeck;
begin
  n := CardList.Count;
  if n = 0 then
  begin
    MessageDlg('No hands dealt', mtError, [mbOk], 0);
    Exit;
  end;
  StatsList := TStringList.Create;
  cr1 := RankBox.ItemIndex;
  StatsList.Add('Rank Selected: ' + RankBox.Text);
  StatsList.Add('Shuffles: ' + IntToStr(n));
  StatsList.Add('');
  StatsList.Add('Position = Count  (Percent) Avg = 7.6923%');
  StatsList.Add('');
  for i := 1 to 52 do PosCount[i] := 0;
  for i := 0 to n - 1 do
  begin
    Deck := CardList[i];
    for j := 1 to 52 do
    begin
      cr2 := (Deck^[j] - 1) div 4;
      if cr1 = cr2 then Inc(PosCount[j]);
    end;
  end;
  for i := 1 to 52 do
  begin
    j := PosCount[i];
    pt := FloatToStrF(j * 100 / n, ffFixed, 18, 4);
    StatsList.Add(IntToStr(i) + ' = ' + IntToStr(j) + '  (' + pt + '%)');
  end;
  with StatsForm do
  begin
    Memo1.Lines.Text := StatsList.Text;
    ShowModal;
  end;
  StatsList.Free;
end;

procedure TForm1.CardStatsBtnClick(Sender: TObject);
var
  i, j, n, CardNum: integer;
  pt: string;
  StatsList: TStringList;
  PosCount: array[1..52] of integer;
  Deck: PCardDeck;
begin
  n := CardList.Count;
  if n = 0 then
  begin
    MessageDlg('No hands dealt', mtError, [mbOk], 0);
    Exit;
  end;
  StatsList := TStringList.Create;
  CardNum := CardBox.ItemIndex + 1;
  StatsList.Add('Card Selected: ' + CardBox.Text);
  StatsList.Add('Shuffles: ' + IntToStr(n));
  StatsList.Add('');
  StatsList.Add('Position = Count  (Percent) Avg = 1.9231%');
  StatsList.Add('');
  for j := 1 to 52 do PosCount[j] := 0;
  for i := 0 to n - 1 do
  begin
    Deck := CardList[i];
    for j := 1 to 52 do if CardNum = Deck^[j] then
    begin
      Inc(PosCount[j]);
      Break;
    end;
  end;
  for i := 1 to 52 do
  begin
    j := PosCount[i];
    pt := FloatToStrF(j * 100 / n, ffFixed, 18, 4);
    StatsList.Add(IntToStr(i) + ' = ' + IntToStr(j) + '  (' + pt + '%)');
  end;
  with StatsForm do
  begin
    Memo1.Lines.Text := StatsList.Text;
    ShowModal;
  end;
  StatsList.Free;
end;

procedure TForm1.GetHandCards(i: integer; var H1, H2, B1, B2, B3, B4, B5: integer);
var
  Dealt, Card1, Card2, Card3, Card4, Card5, Card6, Card7: integer;
  Deck: PCardDeck;
begin
  H1 := 0;
  H2 := 0;
  B1 := 0;
  B2 := 0;
  B3 := 0;
  B4 := 0;
  B5 := 0;
  if (i < 0) or (i > CardList.Count - 1) then Exit;
  Deck := CardList[i];
  if PlayerPosBox.ItemIndex >= Players then
    Card1 := Players - ((PlayerPosBox.ItemIndex - Players + i) mod Players)
  else Card1 := Players - PlayerPosBox.ItemIndex;
  Card2 := Card1 + Players;
  Card3 := (Players * 2) + 1;
  Card4 := Card3 + 1;
  Card5 := Card4 + 1;
  Card6 := Card5 + 1;
  Card7 := Card6 + 1;
  Dealt := DealtBox.ItemIndex + 1;
  H1 := Deck^[Card1];
  H2 := Deck^[Card2];
  if Dealt > 1 then     // flop
  begin
    B1 := Deck^[Card3];
    B2 := Deck^[Card4];
    B3 := Deck^[Card5];
  end;
  if Dealt > 2 then B4 := Deck^[Card6];  // turn
  if Dealt > 3 then B5 := Deck^[Card7];  // river
end;

procedure TForm1.ShowHand(Node: PVirtualNode);
var
  i, Dealt: integer;
  HC1, HC2, B1, B2, B3, B4, B5, Rank: integer;
  Hand, Hole, Board: string;
begin
  if Node = Nil then Hand := '' else
  begin
    i := CardView.AbsoluteIndex(Node);
    GetHandCards(i, HC1, HC2, B1, B2, B3, B4, B5);
    Hole := '[' + GetCard(HC1) + ' ' + GetCard(HC2) + ']';
    Board := '';
    Dealt := DealtBox.ItemIndex + 1;
    if Dealt > 1 then Board := GetCard(B1) + ' ' + GetCard(B2) + ' ' + GetCard(B3);
    if Dealt > 2 then Board := Board + ' ' + GetCard(B4);
    if Dealt > 3 then Board := Board + ' ' + GetCard(B5);
    Rank := HandRank(HC1, HC2, B1, B2, B3, B4, B5);
    Hand := Hole;
    if Board <> '' then Hand := Hand + ' [' + Board + ']';
    Hand := Hand + '  =  ' + HandLong(Rank, False);
  end;
  HandPanel.Caption := Hand;
end;

procedure TForm1.CardViewFocusChanged(Sender: TBaseVirtualTree;
  Node: PVirtualNode; Column: TColumnIndex);
begin
  ShowHand(Node);
end;

procedure TForm1.CardViewGetText(Sender: TBaseVirtualTree; Node: PVirtualNode;
  Column: TColumnIndex; TextType: TVSTTextType; var CellText: string);
var
  i, j: integer;
  Deck: PCardDeck;
begin
  i := Node.Index;
  if Column = 0 then CellText := IntToStr(i + 1) else
  begin
    Deck := CardList[i];
    j := Deck^[Column];
    CellText := CardFace[j] + CardSuit[j];
  end;
end;

procedure TForm1.CardViewPaintText(Sender: TBaseVirtualTree;
  const TargetCanvas: TCanvas; Node: PVirtualNode; Column: TColumnIndex;
  TextType: TVSTTextType);
const Board: array[0..3] of integer = (0, 3, 4, 5);  // board cards dealt
var BC, PP, i: integer;
begin
  if Column = 0 then Exit;
  if Column <= Players * 2 then
  begin
    i := Node.Index;
    if PlayerPosBox.ItemIndex >= Players then
      PP := Players - ((PlayerPosBox.ItemIndex - Players + i) mod Players)
    else PP := Players - PlayerPosBox.ItemIndex;
    if (Column = PP) or (Column = PP + Players) then TargetCanvas.Font.Color := clRed;
    Exit;
  end;
  BC := Board[DealtBox.ItemIndex];
  if (Column - Players * 2 - 1) < BC then TargetCanvas.Font.Color := clBlue
  else TargetCanvas.Font.Color := clGray;
end;

end.
